from aws_sagemaker_remote.training.main import sagemaker_training_main
from .args import train_args


def igniter_main(main, training_args=None, **sagemaker_args):
    r"""
    Run a training script

    Parameters
    ----------
    main : function
        Main function. Must have one argument ``parser:argparse.Namespace``.
    training_args : dict
        Keyword arguments to :meth:`pytorch_igniter.args.train_args`.
    sagemaker_args : dict
        Keyword arguments to ``aws-sagemaker-remote.training.main.sagemaker_training_main``.
        See `aws_sagemaker_remote <https://aws-sagemaker-remote.readthedocs.io/en/latest/>`_.
    """
    argparse_callback_input = sagemaker_args.get('argparse_callback', None)
    if training_args is None:
        training_args = {}

    def argparse_callback(parser):
        train_args(parser=parser, **training_args)
        if argparse_callback_input:
            argparse_callback_input(parser)
    sagemaker_args['argparse_callback'] = argparse_callback
    return sagemaker_training_main(
        main=main,
        **sagemaker_args
    )



import argparse
import os
import torch
import distutils
from aws_sagemaker_remote.args import bool_argument

TRAIN_KWARGS = [
    'n_saved',
    'output_dir',
    'model_dir',
    'checkpoint_dir',
    'mlflow_enable',
    'mlflow_tracking_uri',
    'device',
    'max_epochs',
    'mlflow_experiment_name',
    'mlflow_run_name',
    'is_sagemaker',
    'sagemaker_job_name',
    'mlflow_tracking_username',
    'mlflow_tracking_password',
    'mlflow_tracking_secret_name',
    'mlflow_tracking_secret_region',
    'mlflow_tracking_secret_profile'
]


def train_kwargs(args):
    return {k: getattr(args, k) for k in TRAIN_KWARGS if hasattr(args, k)}


def train_args(
    parser: argparse.ArgumentParser,
    batch_size=64,
    eval_batch_size=64,
    max_epochs=100,
    learning_rate=0.0003,
    device=None,
    mlflow_enable=True,
    mlflow_tracking_uri=None,
    mlflow_tracking_password=None,
    mlflow_tracking_username=None,
    mlflow_tracking_secret_name=None,
    mlflow_tracking_secret_region=None,
    mlflow_tracking_secret_profile=None,
    mlflow_experiment_name='default',
    mlflow_run_name=None,
    arguments='all',
    n_saved=5
):
    r"""
    Configure arguments for training

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Argument parser to configure.
    batch_size : int
        Batch size.
    arguments : list
        List of arguments to include (e.g. ``['batch_size``]) or the string  ``"all"`` to include all arguments.
        (default: "all")
    sagemaker_args : dict
        Keyword arguments to ``aws-sagemaker-remote.training.main.sagemaker_training_main``.
        See `aws_sagemaker_remote <https://aws-sagemaker-remote.readthedocs.io/en/latest/>`_.
    """

    if arguments == 'all' or 'workers' in arguments:
        parser.add_argument("--workers", type=int, default=2,
                            help="number of data loading workers (default: 2)")
    if arguments == 'all' or 'batch_size' in arguments:
        parser.add_argument('--batch-size', type=int, default=batch_size, metavar='N',
                            help='input batch size for training (default: {})'.format(batch_size))
    if arguments == 'all' or 'test_batch_size' in arguments:
        parser.add_argument('--test-batch-size', type=int, default=eval_batch_size, metavar='N',
                            help='input batch size for testing (default: {})'.format(eval_batch_size))
    if arguments == 'all' or 'max_epochs' in arguments:
        parser.add_argument('--max-epochs', type=int, default=max_epochs, metavar='N',
                            help='number of epochs to train (default: {})'.format(max_epochs))
    if arguments == 'all' or 'learning_rate' in arguments:
        parser.add_argument('--learning-rate', type=float, default=learning_rate, metavar='LR',
                            help='learning rate (default: {})'.format(learning_rate))
    if arguments == 'all' or 'seed' in arguments:
        parser.add_argument('--seed', type=int, default=1, metavar='S',
                            help='random seed (default: 1)')
    if arguments == 'all' or 'mlflow_enable' in arguments:
        bool_argument(parser, '--mlflow-enable', default=mlflow_enable,
                      help='Enable logging to MLflow (default: {})'.format(mlflow_enable))
    if arguments == 'all' or 'mlflow_experiment_name' in arguments:
        parser.add_argument('--mlflow-experiment-name', default=mlflow_experiment_name,
                            help='Experiment name in MLflow (default: {})'.format(mlflow_experiment_name))
    if arguments == 'all' or 'mlflow_run_name' in arguments:
        parser.add_argument('--mlflow-run-name', default=mlflow_run_name,
                            help='Run name in MLflow (default: {})'.format(mlflow_run_name))
    if arguments == 'all' or 'mlflow_tracking_uri' in arguments:
        parser.add_argument('--mlflow-tracking-uri', default=mlflow_tracking_uri,
                            help='URI of MLflow tracking server (default: ``{}``)'.format(mlflow_tracking_uri))
    if arguments == 'all' or 'mlflow_tracking_username' in arguments:
        parser.add_argument('--mlflow-tracking-username', default=mlflow_tracking_username,type=str,
                            help='Username for MLflow tracking server (default: ``{}``)'.format(mlflow_tracking_username))
    if arguments == 'all' or 'mlflow_tracking_password' in arguments:
        parser.add_argument('--mlflow-tracking-password', default=mlflow_tracking_password,type=str,
                            help='Password for MLflow tracking server (default: ``{}``)'.format(mlflow_tracking_password))
    if arguments == 'all' or 'mlflow_tracking_secret_name' in arguments:
        parser.add_argument('--mlflow-tracking-secret-name', default=mlflow_tracking_secret_name,type=str,
                            help='Secret for accessing MLflow (default: ``{}``)'.format(mlflow_tracking_secret_name))
    if arguments == 'all' or 'mlflow_tracking_secret_profile' in arguments:
        parser.add_argument('--mlflow-tracking-secret-profile', default=mlflow_tracking_secret_profile,type=str,
                            help='Profile for accessing secret for accessing MLflow (default: ``{}``)'.format(mlflow_tracking_secret_profile))
    if arguments == 'all' or 'mlflow_tracking_secret_region' in arguments:
        parser.add_argument('--mlflow-tracking-secret-region', default=mlflow_tracking_secret_region,type=str,
                            help='Region for accessing secret for accessing MLflow (default: ``{}``)'.format(mlflow_tracking_secret_region))
    if arguments == 'all' or 'n_saved' in arguments:
        parser.add_argument('--n-saved', default=n_saved, type=int,
                            help='Number of checkpoints to keep (default: ``{}``)'.format(n_saved))
    if arguments == 'all' or 'device' in arguments:
        if device is None:
            device = "cpu" if not torch.cuda.is_available() else "cuda"
        parser.add_argument("--device", type=str, default=device,
                            help="device to use (default: {})".format(device))

    """
    parser.add_argument('--hosts', type=list,
                        default=json.loads(os.environ.get('SM_HOSTS', '[]')))
    parser.add_argument('--current-host', type=str,
                        default=os.environ.get('SM_CURRENT_HOST', ''))
    parser.add_argument('--num-gpus', type=int,
                        default=os.environ.get('SM_NUM_GPUS', 1))
    parser.add_argument('--log-interval', type=int, default=100, metavar='N',
                        help='how many batches to wait before logging training status')
    parser.add_argument('--backend', type=str, default=None,
                        help='backend for distributed training (tcp, gloo on cpu and gloo, nccl on gpu)')
    """


def parser_for_docs():
    parser = argparse.ArgumentParser()
    train_args(parser=parser)
    from aws_sagemaker_remote.training.args import sagemaker_training_args
    sagemaker_training_args(
        parser=parser,
        script='script.py'
    )
    return parser


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    train_args(
        parser=parser
    )
    parser.parse_args(args=['--help'])



# Use this code snippet in your app.
# If you need more information about configurations or implementing the sample code, visit the AWS docs:
# https://aws.amazon.com/developers/getting-started/python/

import base64
import json


def get_secret(
        profile_name='default',
        secret_name="mlflow-secret",
        region_name=None):
    from botocore.exceptions import ClientError
    import boto3
    if not profile_name:
        profile_name=None
    session = boto3.Session(profile_name=profile_name)
    if not region_name:
        region_name = session.region_name
    client = session.client(
        service_name='secretsmanager',
        region_name=region_name
    )

    # In this sample we only handle the specific exceptions for the 'GetSecretValue' API.
    # See https://docs.aws.amazon.com/secretsmanager/latest/apireference/API_GetSecretValue.html
    # We rethrow the exception by default.

    try:
        get_secret_value_response = client.get_secret_value(
            SecretId=secret_name
        )
    except ClientError as e:
        if e.response['Error']['Code'] == 'DecryptionFailureException':
            # Secrets Manager can't decrypt the protected secret text using the provided KMS key.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
        elif e.response['Error']['Code'] == 'InternalServiceErrorException':
            # An error occurred on the server side.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
        elif e.response['Error']['Code'] == 'InvalidParameterException':
            # You provided an invalid value for a parameter.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
        elif e.response['Error']['Code'] == 'InvalidRequestException':
            # You provided a parameter value that is not valid for the current state of the resource.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
        elif e.response['Error']['Code'] == 'ResourceNotFoundException':
            # We can't find the resource that you asked for.
            # Deal with the exception here, and/or rethrow at your discretion.
            raise e
    else:
        # Decrypts secret using the associated KMS CMK.
        # Depending on whether the secret is a string or binary, one of these fields will be populated.
        if 'SecretString' in get_secret_value_response:
            secret = get_secret_value_response['SecretString']
            secret = json.loads(secret)
            return secret
        else:
            raise NotImplementedError("Secret should be string not binary")
            #secret = base64.b64decode(get_secret_value_response['SecretBinary'])
        # return secret

    # Your code goes here.




from ignite.engine import Events
from .util import TRAIN_MESSAGE, EVAL_MESSAGE


class RunSpec(object):
    @classmethod
    def train_spec(
        cls,
        loader,
        model,
        criteria,
        optimizer,
        preproc=None,
        **kwargs
    ):
        def step(engine, batch):
            model.train()
            model.zero_grad()
            if preproc is not None:
                batch = preproc(batch)
            inputs, targets = batch
            outputs = model(inputs)
            loss = criteria(outputs, targets)
            loss.backward()
            optimizer.step()
            return {
                "loss": loss
            }
        metrics = {'loss': 'loss'}
        return cls(
            loader=loader,
            step=step,
            metrics=metrics,
            **kwargs
        )

    @classmethod
    def eval_spec(
        cls,
        loader,
        model,
        criteria,
        preproc=None,
        **kwargs
    ):
        def step(engine, batch):
            model.eval()
            if preproc is not None:
                batch = preproc(batch)
            inputs, targets = batch
            outputs = model(inputs)
            loss = criteria(outputs, targets)
            return {
                "loss": loss
            }
        metrics = {'loss': 'loss'}
        return cls(
            loader=loader,
            step=step,
            metrics=metrics,
            **kwargs
        )

    def __init__(
        self,
        loader,
        step,
        metrics,
        max_epochs=1,
        epoch_length=None,
        callback=None,
        enable_pbar=True,
        pbar_metrics='all',
        print_event='default',
        print_metrics='all',
        print_fmt='default',
        print_metric_fmt=' | {name}: {value}',
        log_event='default',
        log_metrics='all',
        plot_event='default',
        plot_metrics='all',
        enable_timer=True
    ):
        self.loader = loader
        self.step = step
        self.metrics = metrics
        self.max_epochs = max_epochs
        self.epoch_length = epoch_length
        self.callback = callback
        self.enable_pbar = enable_pbar
        self.pbar_metrics = pbar_metrics
        self.print_event = print_event
        self.print_metrics = print_metrics
        self.print_fmt = print_fmt
        self.print_metric_fmt = print_metric_fmt
        self.log_event = log_event
        self.log_metrics = log_metrics
        self.plot_event = plot_event
        self.plot_metrics = plot_metrics
        self.enable_timer = enable_timer

    def set_defaults(self, is_training=True):
        """
        Fill in the default events for training or evaluation specs
        """
        if self.metrics is None:
            self.metrics = {}
        if self.plot_event == 'default':
            self.plot_event = Events.EPOCH_COMPLETED
        if is_training:
            # Log and print every 100 training iterations
            if self.log_event == 'default':
                self.log_event = Events.ITERATION_COMPLETED(every=100)
            if self.print_event == 'default':
                self.print_event = Events.ITERATION_COMPLETED(every=100)
            if self.print_fmt == 'default':
                self.print_fmt = TRAIN_MESSAGE
        else:
            # Log and print at the end of each evaluation
            if self.log_event == 'default':
                self.log_event = Events.EPOCH_COMPLETED
            if self.print_event == 'default':
                self.print_event = Events.EPOCH_COMPLETED
            if self.print_fmt == 'default':
                self.print_fmt = EVAL_MESSAGE



import argparse
import contextlib
import os
import random
import warnings
import numpy as np
from ignite.contrib.handlers.mlflow_logger import MLflowLogger
import mlflow
from .mlflow_ctx import mlflow_ctx, get_mlflow_logger
import re
import torch
import torch.nn as nn
import torch.optim as optim
from tqdm import tqdm
import torch.utils.data as data

from ignite.engine import Engine, Events
from ignite.handlers import ModelCheckpoint, Timer
from ignite.metrics import RunningAverage
from pytorch_igniter.metrics import SafeAverage

import torchvision.datasets as dset
import torchvision.transforms as transforms
from torch.autograd import backward
import yaml
from .spec import RunSpec
from .engine import build_engine
from .util import handle_exception, get_last_checkpoint, get_metrics, capture_signals
from .ssm import get_secret
LOADED = "Loaded {}, epoch {}, iteration {}"
COMPLETE = "Training complete"


def train(
    to_save,
    train_spec: RunSpec,
    eval_spec: RunSpec,
    eval_event=Events.EPOCH_COMPLETED,
    save_event=Events.EPOCH_COMPLETED,
    n_saved=10,
    mlflow_enable=True,
    mlflow_tracking_uri=None,
    mlflow_tracking_username=None,
    mlflow_tracking_password=None,
    mlflow_tracking_secret_name=None,
    mlflow_tracking_secret_profile=None,
    mlflow_tracking_secret_region=None,
    mlflow_experiment_name=None,
    mlflow_run_name=None,
    model_dir='output',
    checkpoint_dir='output',
    output_dir='output',
    parameters=None,
    device=None,
    max_epochs=None,
    is_sagemaker=False,
    sagemaker_job_name=None
):
    """
    Train a model
    """
    if max_epochs:
        train_spec.max_epochs = max_epochs
    if mlflow_tracking_uri:
        mlflow.set_tracking_uri(mlflow_tracking_uri)
    if mlflow_tracking_username:
        os.environ['MLFLOW_TRACKING_USERNAME'] = mlflow_tracking_username
    if mlflow_tracking_password:
        os.environ['MLFLOW_TRACKING_PASSWORD'] = mlflow_tracking_password
    if mlflow_tracking_secret_name:
        secret = get_secret(
            profile_name=mlflow_tracking_secret_profile,
            secret_name=mlflow_tracking_secret_name,
            region_name=mlflow_tracking_secret_region)
        #print("Secret: {}".format(secret))
        uri = secret.get('uri', None)
        username = secret.get('username', None)
        password = secret.get('password', None)
        if uri:
            #print("Set uri from secret: [{}]".format(uri))
            mlflow.set_tracking_uri(uri)
        if username:
            #print("Set username from secret")
            os.environ['MLFLOW_TRACKING_USERNAME'] = username
        if password:
            #print("Set password from secret")
            os.environ['MLFLOW_TRACKING_PASSWORD'] = password
    if 'MLFLOW_RUN_ID' in os.environ:
        run_id = os.environ['MLFLOW_RUN_ID']
        #output_dir = os.path.join(output_dir, run_id)
        #model_dir = os.path.join(model_dir, run_id)
        #checkpoint_dir = os.path.join(checkpoint_dir, run_id)

    ctx = mlflow_ctx(
        output_dir=output_dir, checkpoint_dir=checkpoint_dir, mlflow_enable=mlflow_enable,
        experiment_name=mlflow_experiment_name, run_name=mlflow_run_name,
        parameters=parameters, is_sagemaker=is_sagemaker, sagemaker_job_name=sagemaker_job_name)
    os.makedirs(output_dir, exist_ok=True)
    os.makedirs(model_dir, exist_ok=True)
    os.makedirs(checkpoint_dir, exist_ok=True)
    with ctx:
        mlflow_logger = get_mlflow_logger(
            output_dir=output_dir,
            checkpoint_dir=checkpoint_dir,
            mlflow_enable=mlflow_enable
        )
        # Create trainer
        trainer = build_engine(
            spec=train_spec,
            output_dir=output_dir,
            mlflow_logger=mlflow_logger,
            tag='train',
            device=device
        )
        to_save = {'trainer': trainer, **to_save}

        # Saver
        checkpoint_handler = ModelCheckpoint(
            checkpoint_dir, filename_prefix="", n_saved=n_saved, require_empty=False)

        def safe_checkpoint_handler(engine, to_save):
            if engine.state.iteration and engine.state.iteration > 0:
                _, last_iteration = get_last_checkpoint(
                    checkpoint_handler=checkpoint_handler)
                if last_iteration is None or last_iteration < engine.state.iteration:
                    checkpoint_handler(engine=engine, to_save=to_save)

        trainer.add_event_handler(
            event_name=save_event,
            handler=safe_checkpoint_handler,
            to_save=to_save
        )

        # Optional evaluation
        if eval_spec is not None:
            assert eval_event is not None
            if not isinstance(eval_spec, dict):
                eval_spec = {
                    'eval': eval_spec
                }
            # Build evaluators
            evaluators = [
                (
                    build_engine(
                        spec=spec,
                        output_dir=output_dir,
                        mlflow_logger=mlflow_logger,
                        tag=tag,
                        trainer=trainer,
                        metric_cls=SafeAverage,
                        is_training=False,
                        device=device
                    ),
                    spec
                )
                for tag, spec in eval_spec.items()
            ]
            # Add evaluation hook to trainer

            def evaluation(engine):
                for evaluator, spec in evaluators:
                    evaluator.run(
                        spec.loader,
                        max_epochs=spec.max_epochs,
                        epoch_length=spec.epoch_length)
            trainer.add_event_handler(
                event_name=eval_event,
                handler=evaluation)

        # Handle ctrl-C or other exceptions

        def exception_callback(engine):
            # Save on exit
            safe_checkpoint_handler(engine=engine, to_save=to_save)
        trainer.add_event_handler(
            event_name=Events.EXCEPTION_RAISED,
            handler=handle_exception,
            callback=exception_callback
        )

        # Get last checkpoint
        checkpoint_file, _ = get_last_checkpoint(checkpoint_handler)
        with capture_signals():
            if checkpoint_file:
                # Load checkpoint
                checkpoint_data = torch.load(checkpoint_file)
                for key, value in to_save.items():
                    value.load_state_dict(checkpoint_data[key])
                tqdm.write(LOADED.format(
                    checkpoint_file, trainer.state.epoch, trainer.state.iteration))
                if Engine._is_done(trainer.state):
                    # Training complete
                    tqdm.write(COMPLETE)
                else:
                    # Continue training
                    trainer.run(train_spec.loader)
            else:
                # Start training
                trainer.run(
                    train_spec.loader,
                    max_epochs=train_spec.max_epochs,
                    epoch_length=train_spec.epoch_length)
        safe_checkpoint_handler(engine=trainer, to_save=to_save)
        if model_dir:
            os.makedirs(model_dir, exist_ok=True)
            torch.save(
                {k: v.state_dict() for k, v in to_save.items()},
                os.path.join(model_dir, 'model.pt')
            )
        return get_metrics(engine=trainer)
